"""
This package includes all general utilities to be used by any kind of class. The utilities are related to the AAS
Archive, Submodel and ConfigMap.
"""

__all__ = ['aas_model_extension_utils', 'properties_file_utils', 'general_utils', 'smia_archive_utils']  # TODO Test to export only a part of the source code in the pypi package